@@ -1,131 +1,262 @@
-use markup5ever_rcdom::Node;
-use markup5ever_rcdom::NodeData;
-use markup5ever_rcdom::RcDom;
+//! Library for sanitizing and converting HTML strings to something that [Pango] can render.
+//!
+//! This library contains several functions to (pre)process text to [Pango Markup].
+//! What to use and when depends on the type of input and the desired result. This can range from
+//! just escaping to converting and sanitizing. See the examples below for what is available
+//! based on the input type.
+//!
+//! The functions below convert strings to strings. If your input can contain several block
+//! elements such as headings, lists, code or quote blocks, see the [`block`] module to convert an
+//! input string into a list of these blocks.
+//!
+//! # Markdown/body HTML
+//!
+//! To handle more HTML, use the [`markup_html`] function. This function supports HTML body markup
+//! such as HTML resulting from a [Markdown]-to-HTML conversion. It tries to convert the input to
+//! [Pango Markup] such that rendering by Pango will make it similar like what a browser would.
+//! This involves adding newlines for paragraphs and lists, converting font styles, etc.
+//!
+//! ```rust
+//! # use html2pango::markup_html;
+//! let m = markup_html("<body>this is some <font color=\"#ff0000\">red text</font>!</body>").unwrap();
+//! assert_eq!(m, "this is some <span foreground=\"#ff0000\">red text</span>!");
+//!
+//! let m = markup_html("<body>a nice <a href=\"https://gnome.org\">link</a>").unwrap();
+//! assert_eq!(m, "a nice <a href=\"https://gnome.org\">link</a>");
+//!
+//! let m = markup_html("<body>some items: <ul><li>first</li><li>second</li></ul><body").unwrap();
+//! assert_eq!(m, "some items: \n • first\n • second\n");
+//! ```
+//!
+//! ## Escaping
+//!
+//! To just escape any HTML reserved characters, use [`html_escape`]:
+//!
+//! ```rust
+//! # use html2pango::html_escape;
+//! let s = html_escape("this is a <tag> & this is \"quoted text\"");
+//! assert_eq!(s, "this is a &lt;tag&gt; &amp; this is &quot;quoted text&quot;");
+//! ```
+//!
+//! ## Matrix custom HTML
+//!
+//! For [Matrix], its [specification][Matrix specification] defines a custom HTML format that
+//! specifies the tags and attributes that can be used. Use [`matrix_html_to_markup`] to handle
+//! this custom HTML input so that input is sanitized before it is converted.
+//!
+//! This function is still work-in-progress!
+//!
+//! # Simple HTML
+//!
+//! By simple HTML, we mean plain text that only contains some formatting tags such as
+//! `<strong>`, `<i>`, `<code>`, etc.
+//! For the full list of supported tags and how they are replaced, see [`markup_from_raw`].
+//!
+//!
+//! ## With sanitization
+//!
+//! If you use [`markup`], supported tags are replaced (if necessary), malformed tags are removed
+//! and HTML reserved characters are escaped.
+//!
+//! ```rust
+//! # use html2pango::markup;
+//! let m = markup("<p><strong>this <i>is &sanitized<f;><unsupported/></i></strong></p>");
+//! assert_eq!(m, "<b>this <i>is &amp;sanitized</i></b>");
+//! ```
+//!
+//! Other unsupported, but valid tags are escaped.
+//!
+//! ```rust
+//! # use html2pango::markup;
+//! let m = markup("this is <span>a tag</span>");
+//! assert_eq!(m, "this is &lt;span&gt;a tag&lt;/span&gt;");
+//! ```
+//!
+//! URIs are replaced by links.
+//!
+//! ```rust
+//! # use html2pango::markup;
+//! let m = markup("go to: https://gnome.org");
+//! assert_eq!(m, "go to: <a href=\"https://gnome.org\">https://gnome.org</a>");
+//! ```
+//!
+//! ## Without sanitization
+//!
+//! Use [`markup_from_raw`] if you have already sanitized input:
+//!
+//! ```rust
+//! # use html2pango::markup_from_raw;
+//! let m = markup_from_raw("<p>this is <unsupported>already sanitized</unsupported></p>");
+//! assert_eq!(m, "this is &lt;unsupported&gt;already sanitized&lt;/unsupported&gt;");
+//! ```
+//!
+//! ## Links
+//!
+//! To just replace URIs by links, use [`markup_links`]:
+//!
+//! ```rust
+//! # use html2pango::markup_links;
+//! let m = markup_links("go to: https://gnome.org");
+//! assert_eq!(m, "go to: <a href=\"https://gnome.org\">https://gnome.org</a>");
+//! ```
+//!
+//! [Markdown]: https://daringfireball.net/projects/markdown/
+//! [Matrix]: https://matrix.org/
+//! [Matrix specification]: https://spec.matrix.org/v1.1/client-server-api/#mroommessage-msgtypes
+//! [Pango]: https://pango.gnome.org
+//! [Pango Markup]: https://docs.gtk.org/Pango/pango_markup.html
+
 use html5ever::tendril::TendrilSink;
 use html5ever::tree_builder::TreeBuilderOpts;
 use html5ever::{parse_document, ParseOpts};
+use markup5ever_rcdom::Node;
+use markup5ever_rcdom::NodeData;
+use markup5ever_rcdom::RcDom;
 
-use lazy_static::lazy_static;
 use linkify::{LinkFinder, LinkKind};
 use maplit::{hashmap, hashset};
+use once_cell::sync::Lazy;
 
-use regex::Regex;
+use regex::{Captures, Regex};
+use std::fmt::Write as _;
 
-const AMP: &str = "(&amp;)";
-const DOMAIN: &str = "[^\\s,)(\"]+";
-const HASH: &str = "(#[\\w._-]+)?";
+use std::borrow::Cow;
 
 pub mod block;
 
-/// Sanitize the input using [`ammonia`][ammonia]'s defaults,
-/// Convert the input `&str` to pango format and parse
-/// URLS to show as `pango` markup links(removes rel attributes).
+/// Sanitizes and converts simple HTML markup to Pango Markup.
 ///
-/// If you want to parse a pre-sanitized input, you can use `markup_from_raw`.
-///
-/// Currently it support conversion of the following `html` tags:
-/// * `<p>` and `</p>` => `""`
-/// * `<i>`, `<em>` and `</i>`,` </em>` => `<i>` and `</i>``
-/// * `<b>`, `<strong>` and `</b>`, `</strong>` => `<b>` and `</b>`
-/// * `<br>` => `\n`
-/// * rest `<` => `&lt`
-/// * rest `>` => `&gt`
-/// * `&nbsp;` => ` ` (non breaking space)
+/// This will sanitize the input string using the [`ammonia`][ammonia] crate's defaults and then
+/// use [`markup_from_raw`] to convert the sanitized string to [Pango Markup].
 ///
 /// # Examples
 ///
 /// ```rust
 /// # use html2pango::markup;
-///
 /// let m = markup("this is parsed");
-/// assert_eq!(&m, "this is parsed");
+/// assert_eq!(m, "this is parsed");
 ///
 /// let m = markup("<b>this <i>is &ssd<f;</i></b>");
-/// assert_eq!(&m, "<b>this <i>is &amp;ssd</i></b>");
+/// assert_eq!(m, "<b>this <i>is &amp;ssd</i></b>");
 ///
 /// let m = markup("this is <span>parsed</span>");
-/// assert_eq!(&m, "this is &lt;span&gt;parsed&lt;/span&gt;");
+/// assert_eq!(m, "this is &lt;span&gt;parsed&lt;/span&gt;");
 ///
 /// let m = markup("with links: http://gnome.org");
-/// assert_eq!(&m, "with links: <a href=\"http://gnome.org\">http://gnome.org</a>");
+/// assert_eq!(m, "with links: <a href=\"http://gnome.org\">http://gnome.org</a>");
 /// ```
 ///
-/// [ammonia]: https://docs.rs/ammonia/1.1.0/ammonia/fn.clean.html
+/// [Pango Markup]: https://docs.gtk.org/Pango/pango_markup.html
+/// [ammonia]: https://docs.rs/ammonia/3.1.2/ammonia/fn.clean.html
 pub fn markup(s: &str) -> String {
     let sanitized_html = ammonia::Builder::new().link_rel(None).clean(s).to_string();
     markup_from_raw(&sanitized_html)
 }
 
-/// Same as `markup` but without sanizing the input.
+/// Converts simple HTML markup to Pango Markup.
+///
+/// This will convert the input string to [Pango Markup] and replace URIs by links.
+///
+/// Currently it support conversion of the following HTML tags:
+/// * `<p>` and `</p>` => `""`
+/// * `<i>`, `<em>` and `</i>`,` </em>` => `<i>` and `</i>`
+/// * `<b>`, `<strong>` and `</b>`, `</strong>` => `<b>` and `</b>`
+/// * `<code>` and `</code>` => `<tt>` and `</tt>`
+/// * `<sub>` and `</sub>` => `<sub>` and `</sub>`
+/// * `<sup>` and `</sup>` => `<sup>` and `</sup>`
+/// * `<br>` => `\n`
+///
+/// It will also convert the following characters and HTML entities:
+/// * `<` => `&lt`
+/// * `>` => `&gt`
+/// * `&nbsp;` => ` ` (non breaking space)
 ///
 /// # Examples
 ///
 /// ```
 /// # use html2pango::markup_from_raw;
-///
 /// let m = markup_from_raw("this is parsed");
-/// assert_eq!(&m, "this is parsed");
+/// assert_eq!(m, "this is parsed");
 ///
 /// let m = markup_from_raw("<b>this <i>is &ssd<f;</i></b>");
-/// assert_eq!(&m, "<b>this <i>is &ssd&lt;f;</i></b>");
+/// assert_eq!(m, "<b>this <i>is &ssd&lt;f;</i></b>");
 ///
 /// let m = markup_from_raw("this is <span>parsed</span>");
-/// assert_eq!(&m, "this is &lt;span&gt;parsed&lt;/span&gt;");
+/// assert_eq!(m, "this is &lt;span&gt;parsed&lt;/span&gt;");
 ///
 /// let m = markup_from_raw("with links: http://gnome.org");
-/// assert_eq!(&m, "with links: <a href=\"http://gnome.org\">http://gnome.org</a>");
+/// assert_eq!(m, "with links: <a href=\"http://gnome.org\">http://gnome.org</a>");
 /// ```
+///
+/// [Pango Markup]: https://docs.gtk.org/Pango/pango_markup.html
 pub fn markup_from_raw(s: &str) -> String {
-    lazy_static! {
-        static ref PARAM: String = format!("({amp}?\\w+(=[\\w._-]+)?)", amp = AMP);
-        static ref PARAMS: String = format!("(\\?{param}*)*", param = *PARAM);
-        static ref REURL: String = format!(
-            "(https?://{domain}{params}{hash})",
-            domain = DOMAIN,
-            params = *PARAMS,
-            hash = HASH
-        );
-        static ref RE: Regex = Regex::new(&REURL).unwrap();
-        static ref MATCH: Regex = Regex::new(
-            r"<p>|</p>|<br>|<b>|</b>|<strong>|</strong>|<code>|</code>|<i>|</i>|<em>|</em>|<|>|&nbsp;"
-        ).unwrap();
-    }
+    static TAGS: Lazy<Regex> = Lazy::new(|| {
+        Regex::new(concat![
+            r"<p>|</p>|",
+            r"<br>|",
+            r"<b>|</b>|",
+            r"<strong>|</strong>|",
+            r"<code>|</code>|",
+            r"<i>|</i>|",
+            r"<sub>|</sub>|",
+            r"<sup>|</sup>|",
+            r"<em>|</em>|",
+            r"<|>|",
+            r"&nbsp;|",
+            // Url
+            r"https?://",                       // Protocol
+            r#"[^\s,)("]+"#,                    // Domain
+            r"(\?((&amp;)?\w+(=[\w._-]+)?)*)*", // Parameters
+            r"(#[\w._-]+)?",                    // Hash
+        ])
+        .unwrap()
+    });
 
-    let s = s.trim();
-
-    let mut previous_end = 0;
-    let mut foo = Vec::with_capacity(s.as_bytes().len());
-    for mat in MATCH.find_iter(s) {
-        foo.push(&s[previous_end..mat.start()]);
-        // Kepp in sync with the regex query or it will panic
-        foo.push(match mat.as_str() {
-            "<p>" | "</p>" => "",
-            "<br>" => "\n",
-            "<b>" | "<strong>" => "<b>",
-            "</b>" | "</strong>" => "</b>",
-            "<code>" => "<tt>",
-            "</code>" => "</tt>",
-            "<i>" | "<em>" => "<i>",
-            "</i>" | "</em>" => "</i>",
-            "<" => "&lt;",
-            ">" => "&gt;",
-            "&nbsp;" => " ",
-            _ => unreachable!(),
-        });
-        previous_end = mat.end();
-    }
-
-    foo.push(&s[previous_end..]);
-    let mut out = String::with_capacity(foo.iter().map(|s| s.len()).sum());
-    out.extend(foo.into_iter());
-
-    String::from(RE.replace_all(&out.trim(), "<a href=\"$0\">$0</a>"))
+    TAGS.replace_all(s.trim(), |captures: &Captures<'_>| {
+        let m = captures.get(0).unwrap().as_str();
+        // Keep this in sync with the TAGS regex
+        match m {
+            "<p>" | "</p>" => Cow::Borrowed(""),
+            "<br>" => Cow::Borrowed("\n"),
+            "<b>" | "<strong>" => Cow::Borrowed("<b>"),
+            "</b>" | "</strong>" => Cow::Borrowed("</b>"),
+            "<code>" => Cow::Borrowed("<tt>"),
+            "</code>" => Cow::Borrowed("</tt>"),
+            "<i>" | "<em>" => Cow::Borrowed("<i>"),
+            "</i>" | "</em>" => Cow::Borrowed("</i>"),
+            "<sub>" => Cow::Borrowed("<sub>"),
+            "</sub>" => Cow::Borrowed("</sub>"),
+            "<sup>" => Cow::Borrowed("<sup>"),
+            "</sup>" => Cow::Borrowed("</sup>"),
+            "<" => Cow::Borrowed("&lt;"),
+            ">" => Cow::Borrowed("&gt;"),
+            "&nbsp;" => Cow::Borrowed(" "),
+            _ => Cow::Owned(format!(r#"<a href="{}">{}</a>"#, m, m)),
+        }
+    })
+    .into()
 }
 
+/// Sanitizes and convert Matrix custom HTML markup (formatted body) to Pango Markup.
+///
+/// First sanitizes the input string by leaving in only the HTML tags and attributes allowed
+/// by the [Matrix specification] and then uses [`markup_html`] to convert the sanitized string to
+/// [Pango Markup].
+///
+/// # Errors
+///
+/// Returns an error if reading from the strings fails.
+///
+/// # Examples
+///
+/// TODO
+///
+/// [Matrix specification]: https://spec.matrix.org/v1.1/client-server-api/#mroommessage-msgtypes
+/// [Pango Markup]: https://docs.gtk.org/Pango/pango_markup.html
 // WIP: only allow the html subset that matrix uses.
 pub fn matrix_html_to_markup(s: &str) -> String {
     // https://github.com/matrix-org/matrix-react-sdk/blob/4bf5e44b2043bbe95faa66943878acad23dfb823/src/HtmlUtils.js#L178-L184
-    #[cfg_attr(rustfmt, rustfmt_skip)]
+    #[rustfmt::skip]
     let allowed_tags = hashset![
         "font", // custom to matrix for IRC-style font coloring
         "del", // for markdown
@@ -163,7 +294,20 @@ pub fn matrix_html_to_markup(s: &str) -> String {
         .unwrap_or_else(|_| markup_from_raw(&sanitized_html))
 }
 
-/// Escape the html entities of `s`
+/// Escapes the HTML reserved characters.
+///
+/// This replaces the HTML reserved characters by their respective entities in the input string.
+///
+/// # Examples
+///
+/// ```rust
+/// # use html2pango::html_escape;
+/// let s = html_escape("this is a <tag>");
+/// assert_eq!(s, "this is a &lt;tag&gt;");
+///
+/// let s = html_escape("he says: \"this & that\"");
+/// assert_eq!(s, "he says: &quot;this &amp; that&quot;");
+/// ```
 pub fn html_escape(s: &str) -> String {
     s.to_string()
         .replace('&', "&amp;")
@@ -172,7 +316,18 @@ pub fn html_escape(s: &str) -> String {
         .replace('"', "&quot;")
 }
 
-/// Converts links to <a href="LINK">LINK</a>
+/// Replaces URIs by HTML link tags.
+///
+/// This function assumes that HTML entities in input string have been escaped (see also:
+/// [`html_escape`]). It will leave URIs in already existing links alone.
+///
+/// # Examples
+///
+/// ```rust
+/// # use html2pango::{html_escape, markup_links};
+/// let m = markup_links("go to https://gnome.org");
+/// assert_eq!(m, "go to <a href=\"https://gnome.org\">https://gnome.org</a>");
+/// ```
 pub fn markup_links(s: &str) -> String {
     let mut parsed = String::with_capacity(s.len());
     let finder = LinkFinder::new();
@@ -208,7 +363,7 @@ pub fn markup_links(s: &str) -> String {
                     let t = s.len() - 6;
                     s.truncate(t);
                 }
-                parsed.push_str(&format!("<a href=\"{0}\">{0}</a>", s))
+                let _ = write!(parsed, "<a href=\"{0}\">{0}</a>", s);
             }
             _ => {
                 if let Some(s) = prepend_str {
@@ -223,7 +378,7 @@ pub fn markup_links(s: &str) -> String {
     parsed
 }
 
-fn convert_tag<'a>(t: &'a str) -> Option<(&'a str, &'a str)> {
+fn convert_tag(t: &str) -> Option<(&str, &str)> {
     let allowed = [
         "a",
         "br",
@@ -235,6 +390,8 @@ fn convert_tag<'a>(t: &'a str) -> Option<(&'a str, &'a str)> {
         "b",
         "blockquote",
         "li",
+        "sub",
+        "sup",
     ];
 
     if !allowed.contains(&t) {
@@ -249,6 +406,8 @@ fn convert_tag<'a>(t: &'a str) -> Option<(&'a str, &'a str)> {
         "code" => Some(("<tt>", "</tt>")),
         "strong" | "b" => Some(("<b>", "</b>")),
         "li" => Some(("", "\n")),
+        "sub" => Some(("<sub>", "</sub>")),
+        "sup" => Some(("<sup>", "</sup>")),
         _ => None,
     }
 }
@@ -268,11 +427,11 @@ fn convert_node(node: &Node, autolinks: bool) -> String {
 
     match node.data {
         NodeData::Text { contents: ref c } => {
-            let escaped = &html_escape(&c.borrow().replace("\n", ""));
+            let escaped = &html_escape(&c.borrow().replace('\n', ""));
             if autolinks {
                 output.push_str(&markup_links(escaped));
             } else {
-                output.push_str(&escaped);
+                output.push_str(escaped);
             }
         }
         NodeData::Element {
@@ -285,23 +444,19 @@ fn convert_node(node: &Node, autolinks: bool) -> String {
 
             match &tag[..] {
                 "ul" => {
-                    content.push_str("\n");
+                    content.push('\n');
                     for child in node.children.borrow().iter() {
                         if match_tag(child, &["li", "ul", "ol"]) {
-                            content.push_str(&format!(" • {}", &convert_node(child, true)));
+                            let _ = write!(content, " • {}", &convert_node(child, true));
                         }
                     }
                 }
                 "ol" => {
                     let mut counter = 1;
-                    content.push_str("\n");
+                    content.push('\n');
                     for child in node.children.borrow().iter() {
                         if match_tag(child, &["li", "ul", "ol"]) {
-                            content.push_str(&format!(
-                                " {}. {}",
-                                counter.to_string(),
-                                &convert_node(child, true)
-                            ));
+                            let _ = write!(content, " {}. {}", counter, &convert_node(child, true));
                             counter += 1;
                         }
                     }
@@ -318,14 +473,11 @@ fn convert_node(node: &Node, autolinks: bool) -> String {
                     output.push_str(&content);
                 }
                 "a" => {
-                    let mut link = "".to_string();
-                    for attr in a.borrow().iter() {
-                        let s = attr.name.local.to_string();
-                        match &s[..] {
-                            "href" => {
-                                link = attr.value.to_string();
-                            }
-                            _ => {}
+                    let mut link = String::new();
+                    let attrs = a.borrow();
+                    for attr in attrs.iter() {
+                        if &*attr.name.local == "href" {
+                            link = attr.value.to_string();
                         }
                     }
 
@@ -334,29 +486,44 @@ fn convert_node(node: &Node, autolinks: bool) -> String {
                         no_link_content.push_str(&convert_node(child, false));
                     }
 
-                    output.push_str(&format!("<a href=\"{}\">{}</a>", html_escape(&link), no_link_content));
+                    let _ = write!(
+                        output,
+                        "<a href=\"{}\">{}</a>",
+                        html_escape(&link),
+                        no_link_content
+                    );
                 }
                 "font" => {
-                    let mut color = "".to_string();
-                    for attr in a.borrow().iter() {
-                        let s = attr.name.local.to_string();
-                        match &s[..] {
-                            "color" => {
-                                color = attr.value.to_string();
-                            }
-                            _ => {}
-                        }
-                    }
-                    output.push_str(&format!(
-                        "<span foreground=\"{}\">{1}</span>",
-                        color, content
-                    ));
+                    let attrs = a.borrow();
+                    let span_attrs = attrs
+                        .iter()
+                        .flat_map(|attr| match &*attr.name.local {
+                            "color" | "data-mx-color" => Some(("foreground", &*attr.value)),
+                            "data-mx-bg-color" => Some(("background", &*attr.value)),
+                            _ => None,
+                        })
+                        .collect::<Vec<(&str, &str)>>();
+
+                    format_span(&mut output, content, span_attrs);
+                }
+                "span" => {
+                    let attrs = a.borrow();
+                    let span_attrs = attrs
+                        .iter()
+                        .flat_map(|attr| match &*attr.name.local {
+                            "data-mx-color" => Some(("foreground", &*attr.value)),
+                            "data-mx-bg-color" => Some(("background", &*attr.value)),
+                            _ => None,
+                        })
+                        .collect::<Vec<(&str, &str)>>();
+
+                    format_span(&mut output, content, span_attrs);
                 }
                 _ => {
                     if let Some((t1, t2)) = convert_tag(&tag) {
-                        output.push_str(&format!("{}{}{}", t1, content, t2));
+                        let _ = write!(output, "{}{}{}", t1, content, t2);
                     } else {
-                        output.push_str(&format!("<{0}>{1}</{0}>", tag, content));
+                        let _ = write!(output, "<{0}>{1}</{0}>", tag, content);
                     }
                 }
             };
@@ -367,6 +534,44 @@ fn convert_node(node: &Node, autolinks: bool) -> String {
     output
 }
 
+fn format_span(output: &mut String, content: String, attrs: Vec<(&str, &str)>) {
+    if attrs.is_empty() {
+        output.push_str(&content);
+        return;
+    }
+
+    let _ = write!(output, "<span");
+    for (attr, val) in attrs.into_iter() {
+        let _ = write!(output, " {}=\"{}\"", attr, val);
+    }
+    let _ = write!(output, ">{}</span>", content);
+}
+
+/// Converts HTML body markup to Pango Markup.
+///
+/// While [`markup`]/[`markup_from_raw`] only supports simple HTML tags such as `<p>`, `<b>` and
+/// `<br>`, this function tries to convert HTML body markup so that the resulting [Pango Markup]
+/// will render to something that is similar to what a browser would render.
+///
+/// # Errors
+///
+/// Returns an error if reading from the strings fails.
+///
+/// # Examples
+///
+/// ```rust
+/// # use html2pango::markup_html;
+/// let m = markup_html("<body>this is some <font color=\"#ff0000\">red text</font>!</body>").unwrap();
+/// assert_eq!(m, "this is some <span foreground=\"#ff0000\">red text</span>!");
+///
+/// let m = markup_html("<body>a nice <a href=\"https://gnome.org\">link</a>").unwrap();
+/// assert_eq!(m, "a nice <a href=\"https://gnome.org\">link</a>");
+///
+/// let m = markup_html("<body>some items: <ul><li>first</li><li>second</li></ul><body").unwrap();
+/// assert_eq!(m, "some items: \n • first\n • second\n");
+/// ```
+///
+/// [Pango Markup]: https://docs.gtk.org/Pango/pango_markup.html
 pub fn markup_html(s: &str) -> Result<String, anyhow::Error> {
     let opts = ParseOpts {
         tree_builder: TreeBuilderOpts {
@@ -460,9 +665,25 @@ mod test {
         let m = markup("hello <font color=\"#112233\">world</font>");
         assert_eq!(&m, "hello <span foreground=\"#112233\">world</span>");
 
+        let m = markup(
+            "hello <font data-mx-color=\"#112233\" data-mx-bg-color=\"#332211\">world</font>",
+        );
+        assert_eq!(
+            &m,
+            "hello <span foreground=\"#112233\" background=\"#332211\">world</span>"
+        );
+
         let m = markup("hello <em><font color=\"#112233\">http://gnome.org</font></em>");
         assert_eq!(&m, "hello <i><span foreground=\"#112233\"><a href=\"http://gnome.org\">http://gnome.org</a></span></i>");
 
+        let m = markup(
+            "added <span data-mx-color=\"#112233\" data-mx-bg-color=\"#332211\">Label</span>",
+        );
+        assert_eq!(
+            &m,
+            "added <span foreground=\"#112233\" background=\"#332211\">Label</span>"
+        );
+
         let url = "http://url.com/test?foo1&foo2=test&foo3#hashing";
         let m = markup(&format!("this is &ssdf; {}", url));
         assert_eq!(
@@ -567,4 +788,24 @@ mod test {
             assert_eq!(&m, e);
         }
     }
+
+    #[test]
+    fn test_superscript() {
+        let strings = [("Une 2<sup>e</sup> entente", "Une 2<sup>e</sup> entente")];
+
+        for &(s, e) in strings.iter() {
+            let m = markup(s);
+            assert_eq!(&m, e);
+        }
+    }
+
+    #[test]
+    fn test_subscript() {
+        let strings = [("H<sub>2</sub>O", "H<sub>2</sub>O")];
+
+        for &(s, e) in strings.iter() {
+            let m = markup(s);
+            assert_eq!(&m, e);
+        }
+    }
 }
